# Copyright 2009 Canonical Ltd.
#
# This file is part of desktopcouch.
#
#  desktopcouch is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# desktopcouch is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with desktopcouch.  If not, see <http://www.gnu.org/licenses/>.
"""Communicate with DBUS and also the APIs it proxies, like Zeroconf."""

import logging

import dbus
from dbus.mainloop.glib import DBusGMainLoop
import avahi
DBusGMainLoop(set_as_default=True)

discovery_service_type = "_couchdb_pairing_invitations._tcp"


def get_local_hostname():
    """Get the name of this host, as Unicode host and domain parts."""
    bus, server = get_dbus_bus_server()
    return server.GetHostName(), server.GetDomainName()

def get_remote_hostname(ip_address_as_ascii, iface=avahi.IF_UNSPEC):
    """Look up the name of a host by its address.  This is like DNS
    PTR in-arpa, but all zeroconf-y."""
    bus, server = get_dbus_bus_server()
    try:
        interface, protocol, aprotocol, address, hostname, flags = \
                server.ResolveAddress(iface, avahi.PROTO_UNSPEC, 
                        ip_address_as_ascii, dbus.UInt32(0))
    except dbus.exceptions.DBusException:
        return ip_address_as_ascii

    return hostname

def get_dbus_bus_server():
    """Common sequence of steps to get a Bus and Server object from DBUS."""
    bus = dbus.SystemBus()
    root_name = bus.get_object(avahi.DBUS_NAME, avahi.DBUS_PATH_SERVER)
    server = dbus.Interface(root_name, avahi.DBUS_INTERFACE_SERVER)
    return bus, server


class Advertisement(object):
    """Represents an advertised service that exists on this host."""

    def __init__(self, port, name="rlx!", stype=discovery_service_type,
                 domain="", host="", text="(unknown)"):
        
        self.logging = logging.getLogger(self.__class__.__name__)

        super(Advertisement, self).__init__()
        self.name = name
        self.stype = stype
        self.domain = domain
        self.host = host
        self.port = port
        if hasattr(text, "keys"):
            self.text = avahi.dict_to_txt_array(text)
        else:
            self.text = text

        self.group = None

        self.publish()
    
    def publish(self):
        """Start the advertisement."""
        bus, server = get_dbus_bus_server()

        entry_group = bus.get_object(avahi.DBUS_NAME, server.EntryGroupNew())
        g = dbus.Interface(entry_group, avahi.DBUS_INTERFACE_ENTRY_GROUP)

        g.AddService(avahi.IF_UNSPEC, avahi.PROTO_UNSPEC, dbus.UInt32(0),
                self.name, self.stype, self.domain, self.host,
                dbus.UInt16(self.port), self.text)

        g.Commit()
        self.logging.info("starting advertising %s on port %d",
                discovery_service_type, self.port)
        self.group = g
    
    def unpublish(self):
        """End the advertisement."""
        self.group.Reset()
        self.logging.info("ending advertising %s on port %d",
                discovery_service_type, self.port)
        self.group = None

    def die(self):
        """Quit."""
        self.unpublish()


def discover_services(add_commport_name_cb, del_commport_name_cb,
        show_local=False):
    """Start looking for services.  Use two callbacks to handle seeing
    new services and seeing services disappear."""

    def remove_item_handler(interface, protocol, name, stype, domain, flags):
        """A service disappeared."""

        if not show_local and flags & avahi.LOOKUP_RESULT_LOCAL:
            return

        del_commport_name_cb(name)

    def new_item_handler(interface, protocol, name, stype, domain, flags):
        """A service appeared."""

        def handle_error(*args):
            """An error in resolving a new service."""
            logging.error("zeroconf ItemNew error for services, %s", args)

        def handle_resolved(*args):
            """Successfully resolved a new service, which we decode and send
            back to our calling environment with the callback function."""
            text = {}
            for l in args[9]:
                k, v = "".join(chr(i) for i in l).split("=", 1)
                text[k] = v

            add_commport_name_cb(args[2], text.get("description", "?"),
                    args[5], args[8], text.get("version", None))

        if not show_local and flags & avahi.LOOKUP_RESULT_LOCAL:
            return

        server.ResolveService(interface, protocol, name, stype, 
            domain, avahi.PROTO_UNSPEC, dbus.UInt32(0), 
            reply_handler=handle_resolved, error_handler=handle_error)


    bus, server = get_dbus_bus_server()

    domain_name = get_local_hostname()[1]

    browser = server.ServiceBrowserNew(avahi.IF_UNSPEC, avahi.PROTO_UNSPEC,
            discovery_service_type, domain_name, dbus.UInt32(0))
    browser_name = bus.get_object(avahi.DBUS_NAME, browser)

    sbrowser = dbus.Interface(browser_name,
            avahi.DBUS_INTERFACE_SERVICE_BROWSER)
    sbrowser.connect_to_signal("ItemNew", new_item_handler)
    sbrowser.connect_to_signal("ItemRemove", remove_item_handler)
